"""Provides an example how to use general layer"""

import sys
import asyncio

from science_mode_4 import SerialPortConnection
from science_mode_4 import DeviceP24
from examples.utils.example_utils import ExampleUtils

async def main() -> int:
    """Main function"""

    # get comport from command line argument
    com_port = ExampleUtils.get_comport_from_commandline_argument()
    # create serial port connection
    connection = SerialPortConnection(com_port)
    # open connection, now we can read and write data
    connection.open()

    # create science mode device
    device = DeviceP24(connection)
    # call initialize to get basic information (serial, versions) and stop any active stimulation/measurement
    # to have a defined state
    await device.initialize()

    # get general layer to call general commands
    general = device.get_layer_general()
    print(f"Device id: {general.device_id}")
    print(f"Firmware version: {general.firmware_version}")
    print(f"Science mode version: {general.science_mode_version}")

    # close serial port connection
    connection.close()
    return 0


if __name__ == "__main__":
    res = asyncio.run(main())
    sys.exit(res)



"""Provides a base class for a connection"""

from abc import ABC, abstractmethod

from .logger import logger


class Connection(ABC):
    """Abstract base class for connection"""


    @abstractmethod
    def open(self):
        """Open connection"""


    @abstractmethod
    def close(self):
        """Close connection"""


    @abstractmethod
    def is_open(self) -> bool:
        """Checks if connection is open"""


    def write(self, data: bytes):
        """Write data to connection"""
        logger().debug("Outgoing data, length: %d, bytes: %s", len(data), data.hex(" ").upper())


    def read(self) -> bytes:
        """Read all data from connection"""
        result = self._read_intern()
        if len(result) > 0:
            logger().debug("Incoming data, length: %d, bytes: %s", len(result), result.hex(" ").upper())
        return result


    @abstractmethod
    def clear_buffer(self):
        """Clear buffer from connection"""


    @abstractmethod
    def _read_intern(self):
        """Read all data from connection"""



"""Provides class for ADS129x chip channel settings register"""

from dataclasses import dataclass
from enum import IntEnum


class Ads129xChannelPowerMode(IntEnum):
    """Represent ADS129xtype for power mode (channel settings register, 0x05 to 0x0C)"""
    NORMAL_OPERATION = 0
    CHANNEL_POWER_DOWN = 1


class Ads129xChannelGain(IntEnum):
    """Represent ADS129xtype for channel gain (channel settings register, 0x05 to 0x0C)"""
    GAIN_1 = 1
    GAIN_2 = 2
    GAIN_3 = 3
    GAIN_4 = 4
    GAIN_6 = 0
    GAIN_8 = 5
    GAIN_12 = 6


class Ads129xChannelInput(IntEnum):
    """Represent ADS129xtype for channel input (channel settings register, 0x05 to 0x0C)"""
    NORMAL_ELECTRODE_INPUT = 0
    INPUT_SHORTED = 1
    CONJUNCTION_WITH_RLD_MEAS = 2 # Used in conjunction with RLD_MEAS bit for RLD measurements
    MVDD_FOR_SUPPLY_MEASUREMENT = 3
    TEMPERATURE_SENSOR = 4
    TEST_SIGNAL = 5
    RLD_DRP = 6 # RLD_DRP (positive electrode is the driver)
    RLD_DRN = 7 # RLD_DRN (negative electrode is the driver)


@dataclass
class Ads129xChannelSettingsRegister:
    """Describes a channel settings register of ADS129x chip"""

    power_mode = Ads129xChannelPowerMode.NORMAL_OPERATION
    gain = Ads129xChannelGain.GAIN_6
    input = Ads129xChannelInput.NORMAL_ELECTRODE_INPUT

    def set_data(self, data: bytes):
        """Convert data to information"""
        tmp = data[0]
        self.power_mode = Ads129xChannelPowerMode((tmp >> 7) & 0x01)
        self.gain = Ads129xChannelGain((tmp >> 4) & 0x07)
        self.input = Ads129xChannelInput((tmp >> 0) & 0x07)


    def get_data(self) -> bytes:
        """Convert information to bytes"""
        tmp = ((self.power_mode << 7) | (self.gain << 4) | (self.input << 0))
        return [tmp]



"""Provides packet classes for dyscom get with type file info"""

from typing import NamedTuple

from science_mode_4.utils.byte_builder import ByteBuilder
from .dyscom_types import DyscomGetType
from .dyscom_helper import DyscomHelper
from .dyscom_get import PacketDyscomGet, PacketDyscomGetAck


class DyscomGetFileInfoResult(NamedTuple):
    """Helper class for dyscom get file info"""
    filename: str
    filesize: int
    checksum: int


class PacketDyscomGetFileInfo(PacketDyscomGet):
    """Packet for dyscom get with type file info"""


    def __init__(self, filename: str = ""):
        super().__init__()
        self._type = DyscomGetType.FILE_INFO
        self._kind = int(self._type)
        self._filename = filename


    def get_data(self) -> bytes:
        bb = ByteBuilder()
        bb.append_bytes(super().get_data())
        bb.append_bytes(DyscomHelper.str_to_bytes(self._filename, 128))
        # maybe more parameters are necessary here
        # file_size, file_checksum
        return bb.get_bytes()


class PacketDyscomGetAckFileInfo(PacketDyscomGetAck):
    """Packet for dyscom get for type file info acknowledge"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._kind = int(DyscomGetType.FILE_INFO)
        self._filename = ""
        self._filesize = 0
        self._checksum = 0

        if not data is None:
            self._filename = DyscomHelper.bytes_to_str(data[2:130], 128)
            self._filesize = int.from_bytes(data[130:134], "big")
            self._checksum = int.from_bytes(data[134:136], "big")


    @property
    def filename(self) -> str:
        """Getter for filename"""
        return self._filename


    @property
    def filesize(self) -> int:
        """Getter for filesize"""
        return self._filesize


    @property
    def checksum(self) -> int:
        """Getter for checksum"""
        return self._checksum



"""Provides low level layer"""

from science_mode_4.protocol.channel_point import ChannelPoint
from science_mode_4.protocol.types import Channel, Connector
from science_mode_4.utils.logger import logger
from science_mode_4.layer import Layer
from .low_level_channel_config import PacketLowLevelChannelConfig
from .low_level_init import PacketLowLevelInit, PacketLowLevelInitAck
from .low_level_stop import PacketLowLevelStop, PacketLowLevelStopAck
from .low_level_types import LowLevelHighVoltageSource, LowLevelMode


class LayerLowLevel(Layer):
    """
    Class for low level layer, uses internally a PacketBuffer to keep track of responses because
    no command waits for response
    """


    async def init(self, mode: LowLevelMode, high_voltage_source: LowLevelHighVoltageSource):
        """Send low level init command and waits for response"""
        p = PacketLowLevelInit()
        p.mode = mode
        p.high_voltage_source = high_voltage_source
        ack: PacketLowLevelInitAck = await self.send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "LowLevelInit")
        logger().info("Low level init")


    async def stop(self):
        """Send low level stop command and waits for response"""
        p = PacketLowLevelStop()
        ack: PacketLowLevelStopAck = await self.send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "LowLevelStop")
        logger().info("Low level stop")


    def send_init(self, mode: LowLevelMode, high_voltage_source: LowLevelHighVoltageSource):
        """Send low level init command"""
        p = PacketLowLevelInit()
        p.mode = mode
        p.high_voltage_source = high_voltage_source
        self.send_packet(p)
        self._packet_buffer.add_open_acknowledge(p)
        logger().info("Low level send init")


    def send_channel_config(self, execute_stimulation: bool, channel: Channel,
                            connector: Connector, points: list[ChannelPoint]):
        """Send low level channel config command"""
        p = PacketLowLevelChannelConfig()
        p.execute_stimulation = execute_stimulation
        p.channel = channel
        p.connector = connector
        p.points = points
        self.send_packet(p)
        self._packet_buffer.add_open_acknowledge(p)
        logger().info("Low level send channel config")


    def send_stop(self):
        """Send low level stop command"""
        p = PacketLowLevelStop()
        self.send_packet(p)
        self._packet_buffer.add_open_acknowledge(p)
        logger().info("Low level send stop")



"""Provides packet classes for dyscom sys"""

from typing import NamedTuple

from science_mode_4.protocol.commands import Commands
from science_mode_4.protocol.types import ResultAndError
from science_mode_4.protocol.packet import Packet, PacketAck
from science_mode_4.utils.byte_builder import ByteBuilder
from .dyscom_types import DyscomSysState, DyscomSysType
from .dyscom_helper import DyscomHelper


class DyscomSysResult(NamedTuple):
    """Helper class for dyscom sys result"""
    sys_type: DyscomSysType
    state: DyscomSysState
    filename: str


class PacketDyscomSys(Packet):
    """Packet for dyscom sys"""


    def __init__(self, sys_type: DyscomSysType = DyscomSysType.UNDEFINED, filename: str = ""):
        super().__init__()
        self._command = Commands.DL_SYS
        self._sys_type = sys_type
        self._filename = filename


    def get_data(self) -> bytes:
        bb = ByteBuilder()
        bb.append_bytes(DyscomHelper.str_to_bytes(self._filename, 128))
        bb.append_byte(self._sys_type)
        return bb.get_bytes()


class PacketDyscomSysAck(PacketAck):
    """Packet for dyscom sys"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._command = Commands.DL_SYS_ACK
        self._result_error = ResultAndError.NO_ERROR
        self._sys_type = DyscomSysType.UNDEFINED
        self._state = DyscomSysState.UNDEFINED
        self._filename = ""

        if not data is None:
            self._result_error = ResultAndError(data[0])
            self._sys_type = DyscomSysType(data[1])
            self._state = DyscomSysState(data[2])
            self._filename = DyscomHelper.bytes_to_str(data[3:131], 128)


    @property
    def result_error(self) -> ResultAndError:
        """Getter for ResultError"""
        return self._result_error


    @property
    def sys_type(self) -> DyscomSysType:
        """Getter for sys_type"""
        return self._sys_type


    @property
    def state(self) -> DyscomSysState:
        """Getter for state"""
        return self._state


    @property
    def filename(self) -> str:
        """Getter for _filename"""
        return self._filename



"""Provides packet classes for dyscom get with type battery status"""

from typing import NamedTuple
import struct

from .dyscom_types import DyscomEnergyFlag, DyscomGetType
from .dyscom_get import PacketDyscomGet, PacketDyscomGetAck


class DyscomGetBatteryResult(NamedTuple):
    """Helper class for dyscom get with type file by name"""
    voltage: int
    current: int
    percentage: int
    temperature: int
    energy_state: set[DyscomEnergyFlag]


class PacketDyscomGetBatteryStatus(PacketDyscomGet):
    """Packet for dyscom get with type battery status"""


    def __init__(self):
        super().__init__()
        self._type = DyscomGetType.BATTERY
        self._kind = int(self._type)


class PacketDyscomGetAckBatteryStatus(PacketDyscomGetAck):
    """Packet for dyscom get for type battery status acknowledge"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._kind = int(DyscomGetType.BATTERY)
        self._voltage = 0
        self._current = 0
        self._percentage = 0
        self._temperature = 0
        self._energy_state: set[DyscomEnergyFlag] = set()

        if not data is None:
            energy_state, self._percentage, self._temperature, self._current, \
                self._voltage = struct.unpack("<BBbiI", data[2:13])

            for f in DyscomEnergyFlag:
                if energy_state & (1 << f) == 1:
                    self._energy_state.add(f)


    @property
    def voltage(self) -> int:
        """Getter for voltage, [0, 65535] in millivolt"""
        return self._voltage


    @property
    def current(self) -> int:
        """Getter for current, [-327675, 327675] in milliampere"""
        return self._current


    @property
    def percentage(self) -> int:
        """Getter for percentage, [0, 100] in percent"""
        return self._percentage


    @property
    def temperature(self) -> int:
        """Getter for temperature, [-128, 127] in degrees"""
        return self._temperature


    @property
    def energy_state(self) -> set[DyscomEnergyFlag]:
        """Getter for energy state"""
        return self._energy_state



"""Provides a packet number generator"""


class PacketNumberGenerator():
    """Packet number generator class that increases number for each packet by 1"""


    def __init__(self):
        self._current_number = 0


    def get_next_number(self) -> int:
        """Returns next packet number, this function modifies state of class"""
        self._current_number += 1
        # packet number is a 6-bit value
        if self._current_number > 63:
            self._current_number = 0

        return self._current_number



"""Init file"""

from importlib.metadata import PackageNotFoundError, version

from .protocol import *
from .utils import *
from .general import *
from .low_level import *
from .mid_level import *
from .dyscom import *

from .device import *
from .layer import *
from .device_p24 import *
from .device_i24 import *

try:
    __version__ = version("science_mode_4")
    logger().info("Library version %s", __version__)
except PackageNotFoundError:
    pass



"""Provides packet classes for dyscom get with type device id"""

from .dyscom_types import DyscomGetType
from .dyscom_helper import DyscomHelper
from .dyscom_get import PacketDyscomGet, PacketDyscomGetAck


class PacketDyscomGetDeviceId(PacketDyscomGet):
    """Packet for dyscom get with type device id"""


    def __init__(self):
        super().__init__()
        self._type = DyscomGetType.DEVICE_ID
        self._kind = int(self._type)


class PacketDyscomGetAckDeviceId(PacketDyscomGetAck):
    """Packet for dyscom get for type device id acknowledge"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._kind = int(DyscomGetType.DEVICE_ID)
        self._device_id = ""

        if not data is None:
            self._device_id = DyscomHelper.bytes_to_str(data[2:130], 128)


    @property
    def device_id(self) -> str:
        """Getter for device id"""
        return self._device_id



"""Provides packet classes for dyscom send file"""

import struct

from science_mode_4.protocol.commands import Commands
from science_mode_4.protocol.packet import Packet, PacketAck
from science_mode_4.utils.byte_builder import ByteBuilder


class PacketDyscomSendFile(PacketAck):
    """Packet for dyscom send file ack (this is technically not an acknowledge, but it is handled as such,
    because it is send automatically from device). This should probably never be used on PC side"""


    _unpack_func = struct.Struct(">IH").unpack


    def __init__(self, data: bytes):
        super().__init__(data)
        self._command = Commands.DL_SEND_FILE
        self._block_number = 0
        self._block_size = 0
        self._data: bytes = bytes()

        if not data is None:
            self._block_number, self._block_size = PacketDyscomSendFile._unpack_func(">IH", data)
            self._data = data[6:self._block_size]


    @property
    def block_number(self) -> int:
        """Getter for block number"""
        return self._block_number


    @property
    def block_size(self) -> int:
        """Getter for block size"""
        return self._block_size


    @property
    def data(self) -> bytes:
        """Getter for data"""
        return self._data


class PacketDyscomSendFileAck(Packet):
    """Packet for dyscom send file acknowledge (this is technically not a packet, but it is handled as such,
    because it is send from PC to device)"""


    def __init__(self, block_number: int = 0):
        super().__init__()
        self._command = Commands.DL_SEND_FILE_ACK
        self._block_number = block_number


    def get_data(self) -> bytes:
        bb = ByteBuilder()
        bb.append_value(self._block_number, 4, True)
        return bb.get_bytes()



"""Provides packet classes for dyscom power module"""

from typing import NamedTuple
from science_mode_4.protocol.commands import Commands
from science_mode_4.protocol.types import ResultAndError
from science_mode_4.protocol.packet import Packet, PacketAck
from science_mode_4.utils.byte_builder import ByteBuilder
from .dyscom_types import DyscomPowerModuleType, DyscomPowerModulePowerType


class DyscomPowerModuleResult(NamedTuple):
    """Helper class for dyscom power module result"""
    module: DyscomPowerModuleType
    power: DyscomPowerModulePowerType


class PacketDyscomPowerModule(Packet):
    """Packet for dyscom power module"""


    def __init__(self, module: DyscomPowerModuleType = DyscomPowerModuleType.MEASUREMENT, power = DyscomPowerModulePowerType.SWITCH_OFF):
        super().__init__()
        self._command = Commands.DL_POWER_MODULE
        self._module = module
        self._power = power


    def get_data(self) -> bytes:
        bb = ByteBuilder()
        bb.append_byte(self._module)
        bb.append_byte(self._power)
        return bb.get_bytes()


class PacketDyscomPowerModuleAck(PacketAck):
    """Packet for dyscom power module"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._command = Commands.DL_POWER_MODULE_ACK
        self._result_error = ResultAndError.NO_ERROR
        self._module = DyscomPowerModuleType.MEASUREMENT
        self._power = DyscomPowerModulePowerType.SWITCH_OFF

        if not data is None:
            self._result_error = ResultAndError(data[0])
            self._module = DyscomPowerModuleType(data[1])
            self._power = DyscomPowerModulePowerType(data[2])


    @property
    def result_error(self) -> ResultAndError:
        """Getter for ResultError"""
        return self._result_error


    @property
    def module(self) -> DyscomPowerModuleType:
        """Getter for module"""
        return self._module


    @property
    def power(self) -> DyscomPowerModulePowerType:
        """Getter for power"""
        return self._power



"""Provides general layer"""

from science_mode_4.protocol.packet_number_generator import PacketNumberGenerator
from science_mode_4.protocol.packet_factory import PacketFactory
from science_mode_4.layer import Layer
from science_mode_4.utils import logger
from science_mode_4.utils.packet_buffer import PacketBuffer
from .general_reset import PacketGeneralReset, PacketGeneralResetAck
from .general_stim_status import GetStimStatusResult, PacketGeneralGetStimStatus, PacketGeneralGetStimStatusAck
from .general_device_id import PacketGeneralGetDeviceId, PacketGeneralGetDeviceIdAck
from .general_version import PacketGeneralGetExtendedVersion, PacketGeneralGetExtendedVersionAck, GetExtendedVersionResult


class LayerGeneral(Layer):
    """Class for general layer"""


    def __init__(self, packet_buffer: PacketBuffer, packet_factory: PacketFactory, packet_number_generator: PacketNumberGenerator):
        super().__init__(packet_buffer, packet_factory, packet_number_generator)
        self._device_id: str | None = None
        self._firmware_version: str | None = None
        self._science_mode_version: str | None = None


    @property
    def device_id(self) -> str | None:
        """Getter for device id"""
        return self._device_id


    @property
    def firmware_version(self) -> str | None:
        """Getter for firmware version"""
        return self._firmware_version


    @property
    def science_mode_version(self) -> str | None:
        """Getter for science mode version"""
        return self._science_mode_version


    async def initialize(self):
        """Calls initialize commands"""
        await self.get_device_id()
        await self.get_version()


    async def get_device_id(self) -> str:
        """Send get device id command and waits for response"""
        p = PacketGeneralGetDeviceId()
        ack: PacketGeneralGetDeviceIdAck = await self.send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "GetDeviceId")
        self._device_id = ack.device_id
        logger().info("Get device id: %s", ack.device_id)
        return self._device_id


    async def reset(self):
        """Sends reset command and waits for response"""
        logger().info("Reset",)
        p = PacketGeneralReset()
        # maybe we get no ack because the device resets before sending ack
        ack: PacketGeneralResetAck = await self.send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "Reset")


    async def get_stim_status(self) -> GetStimStatusResult:
        """Sends get stim status and waits for response"""
        p = PacketGeneralGetStimStatus()
        ack: PacketGeneralGetStimStatusAck = await self.send_packet_and_wait(p)
        if not ack.successful:
            raise ValueError("Error get stim status")
        logger().info("Get stim status: %s, active: %r", ack.stim_status.name, ack.high_voltage_on)
        return GetStimStatusResult(ack.stim_status, ack.high_voltage_on)


    async def get_version(self) -> GetExtendedVersionResult:
        """Sends get extended version and waits for response, returns firmware and science mode version"""      
        p = PacketGeneralGetExtendedVersion()
        ack: PacketGeneralGetExtendedVersionAck = await self.send_packet_and_wait(p)
        if not ack.successful:
            raise ValueError("Error get extended version")
        self._firmware_version = ack.firmware_version
        self._science_mode_version = ack.science_mode_version
        logger().info("Get version, firmware version: %s science mode version: %s, firmware hash: %s, hash type: %s, is valid hash: %r",\
                      ack.firmware_version, ack.science_mode_version, ack.firmware_hash, ack.hash_type.name, ack.is_valid_hash)
        return GetExtendedVersionResult(self._firmware_version, self._science_mode_version, ack.firmware_hash,\
                                        ack.hash_type, ack.is_valid_hash)



"""Init file for dyscom ADS129x"""


from .ads129x_channel_settings_register import *
from .ads129x_config_register_1 import *
from .ads129x_config_register_2 import *
from .ads129x_config_register_3 import *
from .ads129x_config_register_4 import *
from .ads129x_respiration_control_register import *
from .ads129x import *



"""Test program how to use library without installing the library,
DO NOT USE THIS FILE, USE EXAMPLES INSTEAD"""

import sys
import asyncio

from science_mode_4 import DeviceP24
from science_mode_4.low_level.low_level_channel_config import PacketLowLevelChannelConfigAck
from science_mode_4.protocol import ChannelPoint
from science_mode_4.protocol.commands import Commands
from science_mode_4.utils import SerialPortConnection
from science_mode_4.low_level import LayerLowLevel
from science_mode_4.protocol import Connector, Channel
from science_mode_4.low_level import LowLevelHighVoltageSource, LowLevelMode


async def main() -> int:
    """Main function"""

    # logger().setLevel(logging.DEBUG)
    current = 70

    # # keyboard is our trigger to start specific stimulation
    # def input_callback(input_value: str) -> bool:
    #     """Callback call from keyboard input thread"""
    #     print(f"Input value {input_value}")

    #     nonlocal current
    #     if input_value == "1":
    #         send_channel_config(low_level_layer, Connector.GREEN)
    #     elif input_value == "2":
    #         send_channel_config(low_level_layer, Connector.YELLOW)
    #     elif input_value == "+":
    #         current += 0.5
    #         print(f"current: {current}")
    #     elif input_value == "-":
    #         current -= 0.5
    #         print(f"current: {current}")
    #     elif input_value == "q":
    #         # end keyboard input thread
    #         return True
    #     else:
    #         print("Invalid command")

    #     return False


    def send_channel_config(low_level_layer: LayerLowLevel, connector: Connector):
        """Sends channel update"""
        # device can store up to 10 channel config commands
        for channel in Channel:
            # send_channel_config does not wait for an acknowledge
            low_level_layer.send_channel_config(True, channel, connector,
                                                [ChannelPoint(1000, current), ChannelPoint(4000, 0),
                                                ChannelPoint(1000, -current)])


    print("Usage:")
    print("Press 1 or 2 to stimulate green or yellow connector")
    print("Press + or - to increase or decrease current")
    print("Press q to quit")
    # create keyboard input thread for non blocking console input
    # keyboard_input_thread = KeyboardInputThread(input_callback)

    # get comport from command line argument
    # com_port = ExampleUtils.get_comport_from_commandline_argument()
    # create serial port connection
    connection = SerialPortConnection(SerialPortConnection.list_science_mode_device_ports()[0].device)
    # open connection, now we can read and write data
    connection.open()

    # create science mode device
    device = DeviceP24(connection)
    # call initialize to get basic information (serial, versions) and stop any active stimulation/measurement
    # to have a defined state
    await device.initialize()

    # get low level layer to call low level commands
    low_level_layer = device.get_layer_low_level()

    # call init low level
    await low_level_layer.init(LowLevelMode.NO_MEASUREMENT, LowLevelHighVoltageSource.STANDARD)

    # now we can start stimulation
    # while keyboard_input_thread.is_alive():
    for x in range(500):
        if x % 100 == 0:
            send_channel_config(low_level_layer, Connector.YELLOW)
        # get new packets from connection
        ack = low_level_layer.packet_buffer.get_packet_from_buffer()
        if ack and ack.command == Commands.LOW_LEVEL_CHANNEL_CONFIG_ACK:
            cca: PacketLowLevelChannelConfigAck = ack
            # do something with packet ack
            # here we print that an acknowledge arrived
            print(cca.result.name)

        await asyncio.sleep(0.1)

    # wait until all acknowledges are received
    await asyncio.sleep(0.5)
    # call stop low level
    await low_level_layer.stop()

    # close serial port connection
    connection.close()
    return 0


if __name__ == "__main__":
    res = asyncio.run(main())
    sys.exit(res)



"""Provides a class for a null connection"""


from .connection import Connection


class NullConnection(Connection):
    """Null connection class (only for testing)"""


    def __init__(self):
        self._is_open = False


    def open(self):
        self._is_open = True


    def close(self):
        self._is_open = False


    def is_open(self) -> bool:
        return self._is_open


    def write(self, data: bytes):
        pass


    def read(self) -> bytes:
        result = []
        return bytes(result)



"""Init file for general"""

from .general_device_id import *
from .general_error import *
from .general_layer import *
from .general_reset import *
from .general_stim_status import *
from .general_types import *
from .general_unknown_command import *
from .general_version import *



"""Provides an example how to use mid level layer"""

import sys
import asyncio

from science_mode_4 import DeviceP24
from science_mode_4 import MidLevelChannelConfiguration
from science_mode_4 import ChannelPoint
from science_mode_4 import SerialPortConnection
from examples.utils.example_utils import ExampleUtils


async def main() -> int:
    """Main function"""

    # get comport from command line argument
    com_port = ExampleUtils.get_comport_from_commandline_argument()
    # create serial port connection
    connection = SerialPortConnection(com_port)
    # open connection, now we can read and write data
    connection.open()

    # create science mode device
    device = DeviceP24(connection)
    # call initialize to get basic information (serial, versions) and stop any active stimulation/measurement
    # to have a defined state
    await device.initialize()

    # simple stimulation pattern
    c1p1: ChannelPoint = ChannelPoint(200, 20)
    c1p2: ChannelPoint = ChannelPoint(100, 0)
    c1p3: ChannelPoint = ChannelPoint(200, -20)
    cc1 = MidLevelChannelConfiguration(True, 3, 20, [c1p1, c1p2, c1p3])

    c2p1: ChannelPoint = ChannelPoint(100, 10)
    c2p2: ChannelPoint = ChannelPoint(100, 0)
    c2p3: ChannelPoint = ChannelPoint(100, -10)
    cc2 = MidLevelChannelConfiguration(True, 3, 10, [c2p1, c2p2, c2p3])

    # get mid level layer to call mid level commands
    mid_level = device.get_layer_mid_level()
    # call init mid level, we want to stop on all stimulation errors
    await mid_level.init(True)
    # set stimulation pattern, P24 device will now stimulate according this pattern
    await mid_level.update([cc1, cc2])

    # stimulate for 15s
    for _ in range(15):
        # we have to call get_current_data() every 1.5s to keep stimulation ongoing
        update = await mid_level.get_current_data() # pylint:disable=unused-variable
        # print(update)

        await asyncio.sleep(1)

    # call stop mid level
    await mid_level.stop()

    # close serial port connection
    connection.close()
    return 0


if __name__ == "__main__":
    res = asyncio.run(main())
    sys.exit(res)



"""Provides an example how to use low level layer to measure current and plot values"""

import asyncio
import sys

from science_mode_4 import DeviceP24
from science_mode_4 import ChannelPoint, Commands
from science_mode_4 import SerialPortConnection
from science_mode_4 import PacketLowLevelChannelConfigAck
from science_mode_4 import Connector, Channel
from science_mode_4 import LowLevelHighVoltageSource, LowLevelMode
from science_mode_4 import LayerLowLevel
from examples.utils.example_utils import ExampleUtils
from examples.utils.pyplot_utils import PyPlotHelper


def send_channel_config(low_level_layer: LayerLowLevel):
    """Sends channel update"""
    # device can store up to 10 channel config commands
    for connector in Connector:
        for channel in Channel:
            factor = channel + 1
            # send_channel_config does not wait for an acknowledge
            low_level_layer.send_channel_config(True, channel, connector,
                                                [ChannelPoint(500 * factor, 10 * factor),
                                                 ChannelPoint(500 * factor, 0),
                                                 ChannelPoint(500 * factor, -10 * factor),
                                                 ChannelPoint(500 * factor, 0)])


def calc_plot_index(connector: Connector, channel: Channel) -> int:
    """Calculates index for plot from connector and channel"""
    return connector * len(Channel) + channel


def get_channel_color(channel: Channel) -> str:
    """Retrieves color from channel"""
    color = channel.name
    if color == "WHITE":
        color = "PURPLE"
    return color


async def main() -> int:
    """Main function"""

    plots_info: dict[int, tuple[str, str]] =  {}
    for connector in Connector:
        for channel in Channel:
            plots_info[calc_plot_index(connector, channel)] = f"Connector {connector.name}, channel {channel.name}", get_channel_color(channel)
    plot_helper = PyPlotHelper(plots_info, 500)

    # get comport from command line argument
    com_port = ExampleUtils.get_comport_from_commandline_argument()
    # create serial port connection
    connection = SerialPortConnection(com_port)
    # open connection, now we can read and write data
    connection.open()

    # create science mode device
    device = DeviceP24(connection)
    # call initialize to get basic information (serial, versions) and stop any active stimulation/measurement
    # to have a defined state
    await device.initialize()

    # get low level layer to call low level commands
    low_level_layer = device.get_layer_low_level()

    # call init low level and enable measurement
    await low_level_layer.init(LowLevelMode.STIM_CURRENT, LowLevelHighVoltageSource.STANDARD)

    for _ in range(3):
        # send 8 channel config so we get only 8 acknowledges
        send_channel_config(low_level_layer)

        # wait for stimulation to happen
        await asyncio.sleep(1.0)

        # process all acknowledges and append values to plot data
        while True:
            ack = low_level_layer.packet_buffer.get_packet_from_buffer()
            if ack:
                if ack.command == Commands.LOW_LEVEL_CHANNEL_CONFIG_ACK:
                    ll_config_ack: PacketLowLevelChannelConfigAck = ack
                    # update plot with measured values
                    plot_helper.append_values(calc_plot_index(ll_config_ack.connector, ll_config_ack.channel),
                                              ll_config_ack.measurement_samples)
                    plot_helper.update()
            else:
                break

        await asyncio.sleep(0.1)

    # call stop low level
    await low_level_layer.stop()

    # close serial port connection
    connection.close()

    print("Close plot window to quit")
    plot_helper.loop()
    return 0


if __name__ == "__main__":
    res = asyncio.run(main())
    sys.exit(res)



"""Provides common types"""

from enum import IntEnum


class Channel(IntEnum):
    """Represents a channel"""
    RED = 0
    BLUE = 1
    BLACK = 2
    WHITE = 3


class Connector(IntEnum):
    """Represents a cable connector"""
    YELLOW = 0
    GREEN = 1


class ResultAndError(IntEnum):
    """Represents science mode type ResultAndError"""
    NO_ERROR = 0
    TRANSFER_ERROR = 1
    PARAMETER_ERROR = 2
    PROTOCOL_ERROR = 3
    UC_STIM_TIMEOUT_ERROR = 4
    EMG_TIMEOUT_ERROR = 5
    EMG_REGISTER_ERROR = 6
    NOT_INITIALIZED = 7
    HV_ERROR = 8
    DEMUX_TIMEOUT_ERROR = 9
    ELECTRODE_ERROR = 10
    INVALID_CMD_ERROR = 11
    DEMUX_PARAMETER_ERROR = 12
    DEMUX_NOT_INITIALIZED_ERROR = 13
    DEMUX_TRANSFER_ERROR = 14
    DEMUX_UNKNOWN_ACK_ERROR = 15
    PULSE_TIMEOUT_ERROR = 16
    FUEL_GAUGE_ERROR = 17
    LIVE_SIGNAL_ERROR = 18
    FILE_TRANSMISSION_TIMEOUT = 19
    FILE_NOT_FOUND = 20
    BUSY = 21
    FILE_ERROR = 22
    FLASH_ERASE_ERROR = 23
    FLASH_WRITE_ERROR = 24
    UNKNOWN_CONTROLLER_ERROR = 25
    FIRMWARE_TOO_LARGE_ERROR = 26
    FUEL_GAUGE_NOT_PROGRAMMED = 27
    PULSE_LOW_CURRENT_ERROR = 28


class StimStatus(IntEnum):
    """Represents science mode type StimStatus"""
    NO_LEVEL_INITIALIZED = 0
    LOW_LEVEL_INITIALIZED = 1
    MID_LEVEL_INITIALIZED = 2
    MID_LEVEL_RUNNING = 3



"""Provides base class for packets"""


from .commands import Commands


class Packet():
    """Base class for all packets"""


    def __init__(self):
        self._command = Commands.UNDEFINED
        self._kind = -1
        self._number = 0


    @property
    def command(self) -> Commands:
        """Getter for command"""
        return self._command


    @property
    def kind(self) -> int:
        """Getter for kind (to differentiate between multiple packages per command)"""
        return self._kind


    @property
    def number(self) -> int:
        """Getter for number"""
        return self._number


    @number.setter
    def number(self, value: int):
        """Setter for number"""
        self._number = value


    def get_data(self) -> bytes:
        """Return packet payload"""
        return []


    def create_copy(self) -> "Packet":
        """Returns a copy"""
        return type(self)()


    def __repr__(self) -> str:
        return f"command: {self._command.name}, nr {self._number}"


    def __str__(self) -> str:
        return f"command: {self._command.name}, nr {self._number}"


class PacketAck(Packet):
    """Base class for all acknowledge packets"""


    def __init__(self, data: bytes):
        _ = data
        super().__init__()


    def get_kind(self, data: bytes) -> int:
        """Get kind from data, override in subclasses"""
        _ = data
        return -1


    def create_copy_with_data(self, data: bytes) -> "PacketAck":
        """Returns a copy with data"""
        return type(self)(data)



"""Provides packet classes for dyscom init"""

from typing import NamedTuple
from science_mode_4.dyscom.dyscom_helper import DyscomHelper
from science_mode_4.protocol.commands import Commands
from science_mode_4.protocol.types import ResultAndError
from science_mode_4.protocol.packet import Packet, PacketAck
from .dyscom_types import DyscomFrequencyOut, DyscomInitParams, DyscomInitState
from .ads129x.ads129x import Ads129x


class DyscomInitResult(NamedTuple):
    """Helper class for dyscom get with type file system status"""
    register_map_ads129x: Ads129x
    measurement_file_id: str
    init_state: DyscomInitState
    frequency_out: DyscomFrequencyOut


class PacketDyscomInit(Packet):
    """Packet for dyscom init"""


    def __init__(self, params: DyscomInitParams = DyscomInitParams()):
        super().__init__()
        self._command = Commands.DL_INIT
        self._params = params


    @property
    def params(self) -> DyscomInitParams:
        """Getter for params"""
        return self._params


    @params.setter
    def params(self, value: DyscomInitParams):
        """Setter for params"""
        self._params = value


    def get_data(self) -> bytes:
        return self._params.get_data()


class PacketDyscomInitAck(PacketAck):
    """Packet for dyscom init acknowledge"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._command = Commands.DL_INIT_ACK
        self._result_error = ResultAndError.NO_ERROR
        self._register_map_ads129x: Ads129x
        self._measurement_file_id: str
        self._init_state = DyscomInitState.SUCCESS
        self._frequency_out = DyscomFrequencyOut.SAMPLES_PER_SECOND_4K

        if not data is None:
            self._result_error = ResultAndError(data[0])
            self._register_map_ads129x = Ads129x()
            self._register_map_ads129x.set_data(data[1:27])
            self._measurement_file_id = DyscomHelper.bytes_to_str(data[27:87], 60)
            self._init_state = DyscomInitState(data[87])
            self._frequency_out = DyscomFrequencyOut(data[88])


    @property
    def result_error(self) -> ResultAndError:
        """Getter for ResultError"""
        return self._result_error


    @property
    def register_map_ads129x(self) -> Ads129x:
        """Getter for register map from ADS129x"""
        return self._register_map_ads129x


    @property
    def measurement_file_id(self) -> str:
        """Getter for measurement file id"""
        return self._measurement_file_id


    @property
    def init_state(self) -> DyscomInitState:
        """Getter for init state"""
        return self._init_state


    @property
    def frequency_out(self) -> DyscomFrequencyOut:
        """Getter for frequency out"""
        return self._frequency_out



"""Provides a buffer class to plot values"""

from queue import Empty, Full, Queue
import matplotlib.pyplot as plt
from matplotlib import animation
from matplotlib.backend_bases import CloseEvent

from .plot_base import PlotHelper, PlotValueChannel


class PyPlotValueChannel(PlotValueChannel):
    """Class for holding a specific number of values and handling minimum/maximum specific for PyPlot"""


    def __init__(self, axes: plt.Axes, max_value_count: int, color: str):

        super().__init__(max_value_count)

        # add line with color
        line, = axes.plot(self._x_data, self._y_data, color)
        # enable grid
        axes.grid()

        self._axes = axes
        self._line = line

        self._minimum = None
        self._maximum = None

        # this queue is used to synchronize data between background and main thread
        self._data_queue = Queue(maxsize=0)


    def append_value(self, value: float):
        """Appends a value and updates minimum and maximum for plot. Does run in main thread context"""
        super().append_value(value)

        if self._minimum is None or value < self._minimum:
            self._minimum = value
        if self._maximum is None or value > self._maximum:
            self._maximum = value

        # self._axes.redraw_in_frame()
        try:
            self._data_queue.put_nowait([self._x_data, self._y_data, self._minimum, self._maximum])
        except Full:
            # Queue is full, skip this update
            pass


    def update_plot(self):
        """This function is call in context animation function"""
        super().update_plot()
        try:
            new_x_data, new_y_data, new_minimum, new_maximum = self._data_queue.get_nowait()
            self._line.set_xdata(new_x_data)
            self._line.set_ydata(new_y_data)

            if len(new_x_data) > 1:
                self._axes.set_xlim(new_x_data[0], new_x_data[-1])
                offset = (new_maximum - new_minimum) * 0.1 + 0.1
                self._axes.set_ylim(bottom=new_minimum - offset, top=new_maximum + offset)
        except Empty:
            # No new data in the queue
            pass


class PyPlotHelper(PlotHelper):
    """Class to handle values for plotting"""


    def __init__(self, channels: dict[int, tuple[str, str]], max_value_count: int):
        super().__init__()

        self._window_closed = False
        x_dimension, y_dimension = self._calc_layout_dimension(len(channels))
        self._figure, self._axes = plt.subplots(y_dimension, x_dimension, constrained_layout=True, squeeze=False)
        self._figure.canvas.mpl_connect("close_event", self._on_close)

        for (key, value), sub_plot in zip(channels.items(), self._axes.flat):
            sub_plot.set(xlabel="Samples", ylabel=value[0], title=value[0])
            self._data[key] = PyPlotValueChannel(sub_plot, max_value_count, value[1])

        # interactive mode and show plot
        self._animation_result = animation.FuncAnimation(self._figure, self._animation,
                                                         interval=100, save_count=max_value_count)
        plt.ion()
        plt.show(block=False)
        self.update()


    def update(self):
        """Wait for a short time to process events"""
        plt.pause(0.0001)


    def loop(self):
        """Run event loop until plot window closed"""
        while not self._window_closed:
            self.update()


    def _animation(self, _frame: int, *_fargs: tuple):
        """This function is call in context of main thread"""
        for x in self._data.values():
            x.update_plot()


    def _on_close(self, _event: CloseEvent):
        self._window_closed = True



"""Provides packet classes for dyscom get with type firmware version"""

from .dyscom_types import DyscomGetType
from .dyscom_helper import DyscomHelper
from .dyscom_get import PacketDyscomGet, PacketDyscomGetAck


class PacketDyscomGetFirmwareVersion(PacketDyscomGet):
    """Packet for dyscom get with type firmware version"""


    def __init__(self):
        super().__init__()
        self._type = DyscomGetType.FIRMWARE_VERSION
        self._kind = int(self._type)


class PacketDyscomGetAckFirmwareVersion(PacketDyscomGetAck):
    """Packet for dyscom get for type firmware version acknowledge"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._kind = int(DyscomGetType.FIRMWARE_VERSION)
        self._firmware_version = ""

        if not data is None:
            self._firmware_version = DyscomHelper.bytes_to_str(data[2:130], 128)


    @property
    def firmware_version(self) -> str:
        """Getter for firmware version"""
        return self._firmware_version



"""Provides a buffer class to plot values"""

from queue import Empty, Full, Queue

import numpy as np
import fastplotlib as fpl

from .plot_base import PlotHelper, PlotValueChannel


class FastPlotLibValueChannel(PlotValueChannel):
    """Class for holding a specific number of values specific for FastPlotLib"""


    def __init__(self, sub_plot, max_value_count: int, color: str):
        """sub_plot type is Subplot from fastplotlib"""

        super().__init__(max_value_count)
        self._sub_plot = sub_plot

        # we use an array with zero as start data
        y_data = np.array([0.0] * max_value_count, dtype=np.float32)
        self._line = sub_plot.add_line(y_data, name="values", colors=color)

        # this queue is used to synchronize data between background and main thread
        self._data_queue = Queue(maxsize=0)


    def append_value(self, value: float):
        """Appends a value and returns updated minimum and maximum for plot.
        This function is call in context of background thread"""
        super().append_value(value)

        # because it is not possible to change value count during animation,
        # we fill arrays with "dummy" values
        # for x axis we use values between 0 and 1
        # for y axis we use first real value to fill arrays with that value
        if len(self._x_data) < self._max_value_count:
            self._x_data = [x / self._max_value_count for x in range(self._max_value_count)]
        if len(self._y_data) < self._max_value_count:
            self._y_data = [value] * self._max_value_count

        try:
            self._data_queue.put_nowait([self._x_data, self._y_data])
        except Full:
            # Queue is full, skip this update
            pass


    def update_plot(self):
        """This function is called in context of main thread"""
        super().update_plot()
        try:
            new_x_data, new_y_data = self._data_queue.get_nowait()
            new_x_data = np.array(new_x_data)
            new_y_data = np.array(new_y_data)
            self._line.data[:, 0] = new_x_data
            self._line.data[:, 1] = new_y_data
        except Empty:
            # No new data in the queue
            pass


class FastPlotLibHelper(PlotHelper):
    """Class to handle plots and values for plotting specific for FastPlotLib"""


    def __init__(self, channels: dict[int, tuple[str, str]], max_value_count: int):
        super().__init__()

        # calc layout for sub plots
        x_dimension, y_dimension = self._calc_layout_dimension(len(channels))

        # create figure
        self._figure = fpl.Figure(size=(1024, 768), shape=(y_dimension, x_dimension))

        for (key, value), sub_plot in zip(channels.items(), list(self._figure)):
            sub_plot.title = value[0]
            self._data[key] = FastPlotLibValueChannel(sub_plot, max_value_count, value[1])

        # set animation function that is called regularly to update plots
        self._figure.add_animations(self._animation)
        # show figure
        self._figure.show(maintain_aspect=False)


    def _animation(self, figure: fpl.Figure):
        """This function is called in context of main thread"""
        for x in self._data.values():
            x.update_plot()

        # this keeps the plot in camera view, but prevents any mouse control
        for graph in figure:
            graph.auto_scale()



"""Provides class for ADS129x chip"""

from dataclasses import dataclass

from science_mode_4.utils.byte_builder import ByteBuilder
from .ads129x_config_register_1 import Ads129xConfigRegister1
from .ads129x_config_register_2 import Ads129xConfigRegister2
from .ads129x_config_register_3 import Ads129xConfigRegister3
from .ads129x_config_register_4 import Ads129xConfigRegister4
from .ads129x_respiration_control_register import Ads129xRespirationControlRegister
from .ads129x_channel_settings_register import Ads129xChannelSettingsRegister


@dataclass
class Ads129x:
    """Describes register map of ADS129x chip"""

    # control register
    device_id = 0

    config_register_1 = Ads129xConfigRegister1() # CONFIG1
    config_register_2 = Ads129xConfigRegister2() # CONFIG2
    config_register_3 = Ads129xConfigRegister3() # CONFIG3
    config_register_4 = Ads129xConfigRegister4() # CONFIG4

    lead_off_control_register = 0 # LOFF

    channel_1_setting_register = Ads129xChannelSettingsRegister() # CH1SET
    channel_2_setting_register = Ads129xChannelSettingsRegister() # CH2SET
    channel_3_setting_register = Ads129xChannelSettingsRegister() # CH3SET
    channel_4_setting_register = Ads129xChannelSettingsRegister() # CH4SET

    positive_signal_derivation_register = 0x02 # RLD_SENSP
    negative_signal_derivation_register = 0xEA # RLD_SENSN
    positive_signal_lead_off_detection_register = 0 # LOFF_SENSP
    negative_signal_lead_off_detection_register = 0 # LOFF_SENSN

    lead_off_flip_register = 0 # LOFF_FLIP
    lead_off_positive_signal_status_register = 0 # LOFF_STATP
    lead_off_negative_signal_status_register = 0 # LOFF_STATN
    gpio_register = 0 # GPIO
    pace_detect_register = 0 # PACE

    respiration_control_register = Ads129xRespirationControlRegister() # RESP

    wilson_central_terminal_and_augmented_lead_control_register = 0 # WCT1
    wilson_central_terminal_control_register = 0 # WCT2



    def set_data(self, data: bytes):
        """Convert data to information"""

        self.channel_1_setting_register.set_data([data[0]])
        self.channel_2_setting_register.set_data([data[1]])
        self.channel_3_setting_register.set_data([data[2]])
        self.channel_4_setting_register.set_data([data[3]])

        self.config_register_1.set_data([data[8]])
        self.config_register_2.set_data([data[9]])
        self.config_register_3.set_data([data[10]])
        self.config_register_4.set_data([data[11]])

        self.gpio_register = data[12]
        self.device_id = data[13]
        self.lead_off_control_register = data[14]
        self.lead_off_flip_register = data[15]
        self.negative_signal_lead_off_detection_register = data[16]
        self.positive_signal_lead_off_detection_register = data[17]
        self.lead_off_negative_signal_status_register = data[18]
        self.lead_off_positive_signal_status_register = data[19]

        self.pace_detect_register = data[20]
        self.respiration_control_register.set_data([data[21]])
        self.negative_signal_derivation_register = data[22]
        self.positive_signal_derivation_register = data[23]

        self.wilson_central_terminal_and_augmented_lead_control_register = data[24]
        self.wilson_central_terminal_control_register = data[25]



    def get_data(self) -> bytes:
        """Convert information to bytes"""

        bb = ByteBuilder(0, 26)
        bb.set_bytes_to_position(self.channel_1_setting_register.get_data(), 0, 1)
        bb.set_bytes_to_position(self.channel_2_setting_register.get_data(), 1, 1)
        bb.set_bytes_to_position(self.channel_3_setting_register.get_data(), 2, 1)
        bb.set_bytes_to_position(self.channel_4_setting_register.get_data(), 3, 1)

        bb.set_bytes_to_position(self.config_register_1.get_data(), 8, 1)
        bb.set_bytes_to_position(self.config_register_2.get_data(), 9, 1)
        bb.set_bytes_to_position(self.config_register_3.get_data(), 10, 1)
        bb.set_bytes_to_position(self.config_register_4.get_data(), 11, 1)

        bb.set_bytes_to_position([self.gpio_register], 12, 1)
        bb.set_bytes_to_position([self.device_id], 13, 1)
        bb.set_bytes_to_position([self.lead_off_control_register], 14, 1)
        bb.set_bytes_to_position([self.lead_off_flip_register], 15, 1)
        bb.set_bytes_to_position([self.negative_signal_lead_off_detection_register], 16, 1)
        bb.set_bytes_to_position([self.positive_signal_lead_off_detection_register], 17, 1)
        bb.set_bytes_to_position([self.lead_off_negative_signal_status_register], 18, 1)
        bb.set_bytes_to_position([self.lead_off_positive_signal_status_register], 19, 1)

        bb.set_bytes_to_position([self.pace_detect_register], 20, 1)
        bb.set_bytes_to_position(self.respiration_control_register.get_data(), 21, 1)
        bb.set_bytes_to_position([self.negative_signal_derivation_register], 22, 1)
        bb.set_bytes_to_position([self.positive_signal_derivation_register], 23, 1)

        bb.set_bytes_to_position([self.wilson_central_terminal_and_augmented_lead_control_register], 24, 1)
        bb.set_bytes_to_position([self.wilson_central_terminal_control_register], 25, 1)

        return bb.get_bytes()



"""Provides packet classes for low level channel config"""

from science_mode_4.protocol.commands import Commands
from science_mode_4.protocol.types import Connector, Channel, ResultAndError
from science_mode_4.protocol.packet import Packet, PacketAck
from science_mode_4.protocol.channel_point import ChannelPoint
from science_mode_4.utils.byte_builder import ByteBuilder
from .low_level_types import LowLevelMode


class PacketLowLevelChannelConfig(Packet):
    """Packet for low level channel config"""


    def __init__(self):
        super().__init__()
        self._command = Commands.LOW_LEVEL_CHANNEL_CONFIG
        self._execute_stimulation = False
        self._channel: Channel = Channel.RED
        self._connector: Connector = Connector.YELLOW
        self._points: list[ChannelPoint] = []


    @property
    def execute_stimulation(self) -> bool:
        """Getter for execute stimulation"""
        return self._execute_stimulation


    @execute_stimulation.setter
    def execute_stimulation(self, value: bool):
        """Setter for execute stimulation"""
        self._execute_stimulation = value


    @property
    def channel(self) -> Channel:
        """Getter for channel selection"""
        return self._channel


    @channel.setter
    def channel(self, value: Channel):
        """Setter for channel selection"""
        self._channel = value


    @property
    def connector(self) -> Connector:
        """Getter for connector"""
        return self._connector


    @connector.setter
    def connector(self, value: Connector):
        """Setter for connector"""
        self._connector = value


    @property
    def points(self) -> list[ChannelPoint]:
        """Getter for points"""
        return self._points


    @points.setter
    def points(self, value: list[ChannelPoint]):
        """Setter for points"""
        self._points = value


    def get_data(self) -> bytes:
        if (len(self._points) == 0) or (len(self._points) > 16):
            raise ValueError(f"Low level channel config must have at least 1 point and maximum 16 points {len(self._points)}")

        bb = ByteBuilder()
        bb.set_bit_to_position(len(self._points) - 1, 0, 4)
        bb.set_bit_to_position(self._connector, 4, 1)
        bb.set_bit_to_position(self._channel, 5, 2)
        bb.set_bit_to_position(1 if self._execute_stimulation else 0, 7, 1)
        for x in self._points:
            bb.append_bytes(x.get_data())

        return bb.get_bytes()


class PacketLowLevelChannelConfigAck(PacketAck):
    """Packet for low level channel config acknowledge"""

    def __init__(self, data: bytes):
        super().__init__(data)
        self._command = Commands.LOW_LEVEL_CHANNEL_CONFIG_ACK
        self._result = ResultAndError.NO_ERROR
        self._connector = 0
        self._channel = 0
        self._mode = LowLevelMode.NO_MEASUREMENT
        # only present when measurement is active
        self._sampling_time_in_microseconds: int = 0
        self._measurement_samples: list[float] = None

        if not data is None:
            bb = ByteBuilder()
            bb.append_bytes(data)
            self._result = ResultAndError(data[0])
            self._channel = bb.get_bit_from_position(8, 4)
            self._connector = bb.get_bit_from_position(12, 4)
            self._mode = LowLevelMode(data[2])
            # only present when measurement is active
            if self._mode != LowLevelMode.NO_MEASUREMENT:
                bb.swap(3, 2)
                self._sampling_time_in_microseconds = bb.get_bit_from_position(24, 16)
                self._measurement_samples = [0] * 128
                for index, _ in enumerate(self._measurement_samples):
                    bb.swap(5 + index * 2, 2)
                    self._measurement_samples[index] = bb.get_bit_from_position(40 + index * 16, 16) / 100.0



    @property
    def result(self) -> ResultAndError:
        """Getter for result"""
        return self._result


    @property
    def connector(self) -> int:
        """Getter for connector"""
        return self._connector


    @property
    def channel(self) -> int:
        """Getter for channel"""
        return self._channel


    @property
    def mode(self) -> LowLevelMode:
        """Getter for mode"""
        return self._mode


    @property
    def sampling_time_in_microseconds(self) -> int:
        """Getter for measurement samples"""
        return self._sampling_time_in_microseconds


    @property
    def measurement_samples(self) -> list[float]:
        """Getter for measurement samples,
        by design negative values are measured as positive values,
        unit depends on choosen measurement mode"""
        return self._measurement_samples



"""Provides science mode type channel point"""

from science_mode_4.utils.byte_builder import ByteBuilder


class ChannelPoint():
    """Class for channel point"""


    def __init__(self, duration_micro_seconds: int, current_milli_ampere: float):
        self._duration_in_micro_seconds = duration_micro_seconds
        self._current_in_milli_ampere = current_milli_ampere


    @property
    def current_in_milli_ampere(self) -> float:
        """Getter for current"""
        return self._current_in_milli_ampere


    @property
    def duration_in_micro_seconds(self) -> int:
        """Getter for duration"""
        return self._duration_in_micro_seconds


    def get_data(self) -> bytes:
        """Convert information to bytes"""
        if (self._current_in_milli_ampere < -150.0) or (self._current_in_milli_ampere > 150.0):
            raise ValueError(f"Channel point current must be between -150..150 {self._current_in_milli_ampere}")
        if (self._duration_in_micro_seconds < 0) or (self._duration_in_micro_seconds > 4095):
            raise ValueError(f"Channel point duration must be between 0..4095 {self._duration_in_micro_seconds}")

        c = round(2.0 * self._current_in_milli_ampere + 300.0)

        bb = ByteBuilder()
        bb.set_bit_to_position(0, 0, 10)
        bb.set_bit_to_position(c, 10, 10)
        bb.set_bit_to_position(self._duration_in_micro_seconds, 20, 12)
        bb.swap(0, 4)
        return bb.get_bytes()



"""Provides some helper functionality for examples"""

import sys
import threading
import os
import asyncio
from typing import Callable
from getch import getch

from science_mode_4 import SerialPortConnection


class KeyboardInputThread(threading.Thread):
    """Thread for non blocking keyboard input"""

    def __init__(self, input_cbk: Callable[[str], bool]):
        self._input_cbk = input_cbk
        super().__init__(name = "keyboard_input_thread", daemon = True)
        self.start()


    def run(self):
        while True:
            # getch() returns a bytes object
            key_raw = getch()
            if os.name == "nt":
                key = bytes.decode(key_raw)
            else:
                key = key_raw
            # handle ctrl+c
            if key == "\x03":
                raise KeyboardInterrupt

            if self._input_cbk(key):
                # callback returned True, so end thread
                break


class ExampleUtils():
    """Helper functions for examples"""

    @staticmethod
    def get_comport_from_commandline_argument() -> str:
        """Get com port from command line argument, if no argument provided,
        look for the first matching device,
        if nothing found, end program with exit code 1"""
        if len(sys.argv) != 2:
            # check for available science mode device ports
            ports = SerialPortConnection.list_science_mode_device_ports()
            if len(ports) > 0:
                return ports[0].device

            # nothing found -> exit
            print("No science mode device found")
            print("Serial port command line argument missing (e.g. python -m examples.<layer>.<example> COM3)")
            sys.exit(1)

        com_port = sys.argv[1]
        return com_port


    @staticmethod
    async def wait_for_any_key_pressed(cb: Callable[[], None] | None):
        """Helper function to wait for any key press"""

        # keyboard func
        def input_callback(_input_value: str) -> bool:
            """Callback call from keyboard input thread"""
            # quit on any key
            return True

        # create keyboard input thread for non blocking console input
        keyboard_input_thread = KeyboardInputThread(input_callback)

        # now we can start stimulation
        while keyboard_input_thread.is_alive():
            if cb is not None:
                cb()
            await asyncio.sleep(0.1)



"""Provides science mode mid level types"""

from typing import Sequence
from science_mode_4.protocol.channel_point import ChannelPoint
from science_mode_4.utils.byte_builder import ByteBuilder


class MidLevelChannelConfiguration():
    """Class for mid level channel configuration"""


    def __init__(self, is_active: bool = False, ramp: int = 0, period_in_ms: int = 0, points: Sequence[ChannelPoint] = None):
        self._is_active = is_active
        self._ramp = ramp
        self._period_in_ms = period_in_ms
        self._points: list[ChannelPoint] = [] if points is None else points


    def get_data(self) -> bytes:
        """Returns information as bytes"""
        if len(self._points) > 16:
            raise ValueError(f"Mid level update maximum of 16 points allowed {len(self.points)}")
        if (self._ramp < 0) or (self._ramp > 15):
            raise ValueError(f"Mid level update ramp must be between 0..15 {self.ramp}")
        if (self._period_in_ms < 0) or (self._period_in_ms > 32767 * 4):
            raise ValueError(f"Mid level update period must be between 0..131071 {self._period_in_ms}")

        period_factor = 2 if self._period_in_ms <= 32767 else 4
        bb = ByteBuilder()
        bb.set_bit_to_position(self._ramp, 0, 4)
        bb.set_bit_to_position(len(self._points) - 1, 4, 4)
        bb.set_bit_to_position(0 if period_factor == 2 else 1, 8, 1)
        bb.set_bit_to_position(self._period_in_ms * period_factor, 9, 15)
        bb.swap(1, 2)
        for x in self.points:
            bb.append_bytes(x.get_data())
        return bb.get_bytes()


    @property
    def is_active(self) -> bool:
        """Getter for is active"""
        return self._is_active


    @property
    def ramp(self) -> int:
        """Getter for ramp"""
        return self._ramp


    @property
    def period_in_ms(self) -> int:
        """Getter for period"""
        return self._period_in_ms


    @property
    def points(self) -> list[ChannelPoint]:
        """Getter for points"""
        return self._points



"""Provides an example how to use low level layer"""

import asyncio
import sys

from science_mode_4 import DeviceP24
from science_mode_4 import ChannelPoint
from science_mode_4 import SerialPortConnection
from science_mode_4 import LayerLowLevel
from science_mode_4 import Connector, Channel
from science_mode_4 import LowLevelHighVoltageSource, LowLevelMode
from science_mode_4 import PacketLowLevelChannelConfigAck
from science_mode_4 import Commands
from examples.utils.example_utils import ExampleUtils, KeyboardInputThread


async def main() -> int:
    """Main function"""

    stim_current = 10

    # keyboard is our trigger to start specific stimulation
    def input_callback(input_value: str) -> bool:
        """Callback call from keyboard input thread"""
        print(f"Input value {input_value}")

        nonlocal stim_current
        if input_value == "1":
            send_channel_config(low_level_layer, Connector.GREEN)
        elif input_value == "2":
            send_channel_config(low_level_layer, Connector.YELLOW)
        elif input_value == "+":
            stim_current += 0.5
            print(f"Current: {stim_current}")
        elif input_value == "-":
            stim_current -= 0.5
            print(f"Current: {stim_current}")
        elif input_value == "q":
            # end keyboard input thread
            return True
        else:
            print("Invalid command")

        return False


    def send_channel_config(low_level_layer: LayerLowLevel, connector: Connector):
        """Sends channel update"""
        # device can store up to 10 channel config commands
        for channel in Channel:
            # send_channel_config does not wait for an acknowledge
            low_level_layer.send_channel_config(True, channel, connector,
                                                [ChannelPoint(1000, stim_current), ChannelPoint(4000, 0),
                                                ChannelPoint(1000, -stim_current)])


    print("Usage:")
    print("Press 1 or 2 to stimulate green or yellow connector")
    print("Press + or - to increase or decrease current")
    print("Press q to quit")
    # create keyboard input thread for non blocking console input
    keyboard_input_thread = KeyboardInputThread(input_callback)

    # get comport from command line argument
    com_port = ExampleUtils.get_comport_from_commandline_argument()
    # create serial port connection
    connection = SerialPortConnection(com_port)
    # open connection, now we can read and write data
    connection.open()

    # create science mode device
    device = DeviceP24(connection)
    # call initialize to get basic information (serial, versions) and stop any active stimulation/measurement
    # to have a defined state
    await device.initialize()

    # get low level layer to call low level commands
    low_level_layer = device.get_layer_low_level()

    # call init low level
    await low_level_layer.init(LowLevelMode.NO_MEASUREMENT, LowLevelHighVoltageSource.STANDARD)

    # now we can start stimulation
    while keyboard_input_thread.is_alive():
        # get new packets from connection
        ack = low_level_layer.packet_buffer.get_packet_from_buffer()
        if ack and ack.command == Commands.LOW_LEVEL_CHANNEL_CONFIG_ACK:
            cca: PacketLowLevelChannelConfigAck = ack
            # do something with packet ack
            # here we print that an acknowledge arrived
            print(f"Connector: {cca.connector}, channel: {cca.channel}, result: {cca.result.name}")

        await asyncio.sleep(0.1)

    # wait until all acknowledges are received
    await asyncio.sleep(0.5)
    # call stop low level
    await low_level_layer.stop()

    # close serial port connection
    connection.close()
    return 0


if __name__ == "__main__":
    res = asyncio.run(main())
    sys.exit(res)



"""Provides class for ADS129x chip config register 4"""

from dataclasses import dataclass
from enum import IntEnum


class Ads129xRespirationModulationFrequency(IntEnum):
    """Represent ADS129xtype for respiration modulation frequency (configuration register 4, 0x17)"""
    MODULATION_CLOCK_64KHZ = 0
    MODULATION_CLOCK_32KHZ = 1
    SQUARE_WAVE_16KHZ = 2
    SQUARE_WAVE_8KHZ = 3
    SQUARE_WAVE_4KHZ = 4
    SQUARE_WAVE_2KHZ = 5
    SQUARE_WAVE_1KHZ = 6
    SQUARE_WAVE_500HZ = 7


class Ads129xSingleShotConversion(IntEnum):
    """Represent ADS129xtype for single shot mode (configuration register 4, 0x17)"""
    CONTINUOUS_MODE = 0
    SINGLE_SHOT_MODE = 1


class Ads129xWctToRld(IntEnum):
    """Represent ADS129xtype for WCT to RLD (configuration register 4, 0x17)"""
    CONNECTION_OFF = 0
    CONNECTION_ON = 1


class Ads129xLeadOffComparatorPowerDown(IntEnum):
    """Represent ADS129xtype for lead off comparator power down (configuration register 4, 0x17)"""
    LEAD_OFF_COMPARATOR_DISABLED = 0
    LEAD_OFF_COMPARATOR_ENABLED = 1


@dataclass
class Ads129xConfigRegister4:
    """Describes config register 4 of ADS129x chip"""

    respiration_modulation_frequency = Ads129xRespirationModulationFrequency.MODULATION_CLOCK_64KHZ
    single_shot_conversion = Ads129xSingleShotConversion.CONTINUOUS_MODE
    wct_to_rld = Ads129xWctToRld.CONNECTION_OFF
    lead_off_comparator_power_down = Ads129xLeadOffComparatorPowerDown.LEAD_OFF_COMPARATOR_DISABLED


    def set_data(self, data: bytes):
        """Convert data to information"""
        tmp = data[0]
        self.respiration_modulation_frequency = Ads129xRespirationModulationFrequency((tmp >> 5) & 0x08)
        self.single_shot_conversion = Ads129xSingleShotConversion((tmp >> 3) & 0x01)
        self.wct_to_rld = Ads129xWctToRld((tmp >> 2) & 0x01)
        self.lead_off_comparator_power_down = Ads129xLeadOffComparatorPowerDown((tmp >> 1) & 0x01)


    def get_data(self) -> bytes:
        """Convert information to bytes"""
        tmp = ((self.respiration_modulation_frequency << 5) | (0 << 4) |
               (self.single_shot_conversion << 3) |(self.wct_to_rld << 2) |
               (self.lead_off_comparator_power_down << 1) | (0 << 0))
        return [tmp]



"""Provides a buffer class to plot values"""


class PlotValueChannel:
    """Class for holding a specific number of values and handling minimum/maximum"""


    def __init__(self, max_value_count: int):

        self._x_data: list[float] = []
        self._y_data: list[float] = []

        self._max_value_count = max_value_count
        self._counter = 0


    def append_value(self, value: float):
        """Appends a value and returns updated minimum and maximum for plot"""
        self._x_data.append(self._counter)
        self._counter += 1
        if len(self._x_data) > self._max_value_count:
            self._x_data.pop(0)

        self._y_data.append(value)
        if len(self._y_data) > self._max_value_count:
            self._y_data.pop(0)


    def update_plot(self):
        """Function to update values to plot"""


    @property
    def data(self) -> tuple[list[float], list[float]]:
        """Returns x and y values"""
        return [self._x_data, self._y_data]


class PlotHelper:
    """Class to handle values for plotting"""


    def __init__(self):
        self._data: dict[int, PlotValueChannel] = {}


    @property
    def data(self) -> dict[int, PlotValueChannel]:
        """Returns data with PlotValueChannels"""
        return self._data


    def append_value(self, channel: int, value: float):
        """Append value to channel buffer"""
        self._data[channel].append_value(value)


    def append_values(self, channel: int, values: list[float]):
        """Append values to channel buffer"""
        for x in values:
            self.append_value(channel, x)


    def update(self):
        """Update plot"""


    def loop(self):
        """Run event loop until plot window closed"""


    def _calc_layout_dimension(self, channel_count: int) -> tuple[int, int]:
        """Calculates layout for a specific number of channels, tries to grow
        equal in both directions"""
        layouts = {1: [1, 1], 2: [2, 1], 3: [3, 1], 4: [2, 2], 5: [3, 2], 6: [3, 2], 7: [4, 2], 8: [4, 2]}
        return layouts[channel_count]



"""Provides classes for general GetDeviceId"""

from science_mode_4.protocol.types import ResultAndError
from science_mode_4.protocol.commands import Commands
from science_mode_4.protocol.packet import Packet, PacketAck

class PacketGeneralGetDeviceId(Packet):
    """Packet for general GetDeviceId"""


    def __init__(self):
        super().__init__()
        self._command = Commands.GET_DEVICE_ID


class PacketGeneralGetDeviceIdAck(PacketAck):
    """Packet for general GetDeviceId acknowledge"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._command = Commands.GET_DEVICE_ID_ACK
        self._result_error = ResultAndError.NO_ERROR
        self._device_id = ""

        if not data is None:
            self._result_error = ResultAndError(data[0])
            self._device_id = data[1:11].decode()


    @property
    def result_error(self) -> ResultAndError:
        """Getter for ResultError"""
        return self._result_error


    @property
    def device_id(self) -> str:
        """Getter for DeviceId"""
        return self._device_id



"""Provides helper class for ScienceMode protocol"""

from science_mode_4.utils.byte_builder import ByteBuilder
from science_mode_4.utils.crc16 import Crc16
from science_mode_4.utils import logger
from .packet import Packet


class Protocol:
    """Class for handling ScienceMode protocol packet"""

    START_BYTE = 0xF0
    STOP_BYTE = 0x0F
    STUFFING_BYTE = 0x81
    STUFFING_KEY = 0x55


    @staticmethod
    def packet_to_bytes(packet: Packet) -> bytes:
        """Builds bytes from a packet"""
        # build payload
        bb = ByteBuilder()
        # command and packet number
        bb.set_bit_to_position(packet.command, 0, 10)
        bb.set_bit_to_position(packet.number, 10, 6)
        # swap command and packet number to ensure big endianness
        bb.swap(0, 2)
        # append packet data
        bb.append_bytes(packet.get_data())
        # stuff packet data
        stuffed_packet_data = Protocol.stuff(bb.get_bytes())

        bb.clear()

        # stop byte
        bb.append_byte(Protocol.START_BYTE)
        # packet length
        packet_length: int = len(stuffed_packet_data) + 10
        bb.append_bytes(Protocol.stuff_byte(packet_length >> 8))
        bb.append_bytes(Protocol.stuff_byte(packet_length))
        # crc
        crc_16 = Crc16.crc16_xmodem(stuffed_packet_data)
        bb.append_bytes(Protocol.stuff_byte(crc_16 >> 8))
        bb.append_bytes(Protocol.stuff_byte(crc_16))
        # payload
        bb.append_bytes(stuffed_packet_data)
        # stop byte
        bb.append_byte(Protocol.STOP_BYTE)

        logger().debug("Build package, %s", packet)
        result = bb.get_bytes()
        return bytes(result)


    @staticmethod
    def is_valid_packet_data(buffer: bytes) -> bool:
        """Checks if buffer might contain a valid packet structure, does not check if payload is valid """
        if len(buffer) == 0:
            return False

        result = True

        result &= len(buffer) > 10
        result &= buffer[0] == Protocol.START_BYTE
        result &= buffer[-1] == Protocol.STOP_BYTE
        # packet_length = int.from_bytes([Protocol.unstuffByte(data[2]), Protocol.unstuffByte(data[4])])
        crc = int.from_bytes([Protocol.unstuff_byte(buffer[6]), Protocol.unstuff_byte(buffer[8])])
        result &= crc == Crc16.crc16_xmodem(buffer[9:-1])
        return result


    @staticmethod
    def find_packet_in_buffer(buffer: bytes) -> tuple[int, int] | None:
        """Tries to find a valid packet in buffer, return start and stop index of packet if found or None otherwise"""
        start = 0
        while True:
            # Find start of packet
            # (0xF0 does not always indicate a packet start, so check additionally for stuffing byte)
            start = buffer.find(bytes([Protocol.START_BYTE, Protocol.STUFFING_BYTE]), start)
            if start != -1:
                # we found a start, so use minimal packet length as starting index to find stop
                stop = buffer.find(bytes([Protocol.STOP_BYTE]), start + 12)
                if stop != -1:
                    # we found a packet end, lets check if its valid
                    if Protocol.is_valid_packet_data(buffer[start:stop+1]):
                        return start, stop

                    # found packet is not valid, so check for more packets afterwards
                    start = stop
                else:
                    # we found no stop byte, so there is no complete packet in buffer
                    return None
            else:
                # we found no start byte, so there is no complete packet in buffer
                return None


    @staticmethod
    def extract_packet_data(buffer: bytes) -> tuple[int, int, bytes]:
        """Extract command, packet number and payload from buffer and returns these as tuple, buffer must contain valid packet data"""
        bb = ByteBuilder()
        # command prefix with command and packet number may be stuffed and can be from 2 to 4 bytes long
        command_prefix_count = 0
        for _ in range(2):
            if buffer[9+command_prefix_count] == Protocol.STUFFING_BYTE:
                command_prefix_count += 2
            else:
                command_prefix_count += 1

        bb.append_bytes(Protocol.unstuff(buffer[9:9+command_prefix_count]))
        bb.swap(0, 2)

        command = bb.get_bit_from_position(0, 10)
        nr = bb.get_bit_from_position(10, 6)
        payload = Protocol.unstuff(buffer[9+command_prefix_count:-1])

        return command, nr, payload


    @staticmethod
    def stuff(packet_data: bytes) -> bytes:
        """Stuff data"""
        result: bytearray = bytearray()
        for b in packet_data:
            if b in [Protocol.START_BYTE, Protocol.STOP_BYTE, Protocol.STUFFING_BYTE]:
                result.extend(Protocol.stuff_byte(b))
            else:
                result.append(b)

        return bytes(result)


    @staticmethod
    def unstuff(stuffed_packet_data: bytes) -> bytes:
        """Unstuff data"""
        result = bytearray()
        index: int = 0
        while index < len(stuffed_packet_data):
            if stuffed_packet_data[index] == Protocol.STUFFING_BYTE:
                index += 1
                result.append(Protocol.unstuff_byte(stuffed_packet_data[index]))
            else:
                result.append(stuffed_packet_data[index])

            index += 1

        return bytes(result)


    @staticmethod
    def stuff_byte(b: int) -> bytes:
        """Stuff a single byte"""
        return bytes([Protocol.STUFFING_BYTE, Protocol.STUFFING_KEY ^ (b & 0xFF)])


    @staticmethod
    def unstuff_byte(b: int) -> int:
        """Unstuff a byte, b must be follower of stuffing byte"""
        return Protocol.STUFFING_KEY ^ (b & 0xFF)



"""Provides a class for a usb connection"""

import libusb_package
import usb.core
import usb.util
from .connection import Connection

class UsbConnection(Connection):
    """USB connection class,
    IMPORTANT: work in progress (there are driver issues under windows)"""


    @staticmethod
    def list_devices() -> list[usb.core.Device]:
        """Returns all USB devices"""
        # return list(usb.core.find(find_all=True))
        return libusb_package.find(find_all=True)


    @staticmethod
    def list_science_mode_devices() -> list[usb.core.Device]:
        """Returns all potential science mode USB devices"""
        devices = UsbConnection.list_devices()
        # science mode devices (P24/I24) have an STM32 mcu and these are
        # default values for USB CDC devices
        filtered_devices = list(filter(lambda x: x.idVendor == 0x0483 and x.idProduct == 0x5740 and
                                       x.bDeviceClass == 0x02, devices))
        return filtered_devices


    def __init__(self, device: usb.core.Device):
        self._device = device
        self._out_endpoint: usb.core.Endpoint = None
        self._in_endpoint: usb.core.Endpoint = None
        self._in_endpoint_buffer_size = 128 * 64
        self._is_open = False


    def open(self):
        # P24/I24 have only one configuration
        self._device.set_configuration()
        # get an endpoint instance
        cfg = self._device.get_active_configuration()
        # Interface 0: CDC Communication
        # Interface 1: CDC Data
        interface: usb.core.Interface = cfg[(1,0)]

        # match the first OUT endpoint
        self._out_endpoint = usb.util.find_descriptor(
            interface,
            custom_match = lambda e: usb.util.endpoint_direction(e.bEndpointAddress) == usb.util.ENDPOINT_OUT)
        # match the first IN endpoint
        self._in_endpoint = usb.util.find_descriptor(
            interface,
            custom_match = lambda e: usb.util.endpoint_direction(e.bEndpointAddress) == usb.util.ENDPOINT_IN)

        self._is_open = True


    def close(self):
        self._is_open = False


    def is_open(self) -> bool:
        return self._is_open


    def write(self, data: bytes):
        super().write(data)
        self._out_endpoint.write(data)


    def clear_buffer(self):
        # there is no way to clear the buffer from in endpoint, so
        # read all data and discard it
        self._read_intern()


    def _read_intern(self) -> bytes:
        data = bytes()
        try:
            while True:
                # Read up to endpoint's max packet size (64 bytes in this case)
                tmp = self._in_endpoint.read(self._in_endpoint_buffer_size)
                data += bytes(tmp)
                if len(tmp) < self._in_endpoint_buffer_size:
                    break
        except usb.core.USBError as e:
            if e.errno == 10060:
                # Timeout error (no data available)
                pass
            else:
                raise

        return data
